#include "Energia.h"
#include "lpd8806lib.h"

uint16_t numLEDs = 0;
uint8_t * buffer;  // 1 byte for each B,R,G
uint16_t bufferLen = 0;

// ::: Thanks, advancedLEDbeltKit! :::
// Gamma correction compensates for our eyes' nonlinear perception of
// intensity.  It's the LAST step before a pixel value is stored, and
// allows intermediate rendering/processing to occur in linear space.
// The table contains 256 elements (8 bit input), though the outputs are
// only 7 bits (0 to 127).  This is normal and intentional by design: it
// allows all the rendering code to operate in the more familiar unsigned
// 8-bit colorspace (used in a lot of existing graphics code), and better
// preserves accuracy where repeated color blending operations occur.
// Only the final end product is converted to 7 bits, the native format
// for the LPD8806 LED driver.  Gamma correction and 7-bit decimation
// thus occur in a single operation.
static const uint8_t gammaTable[]  = {
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  1,
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  2,  2,  2,  2,
    2,  2,  2,  2,  2,  3,  3,  3,  3,  3,  3,  3,  3,  4,  4,  4,
    4,  4,  4,  4,  5,  5,  5,  5,  5,  6,  6,  6,  6,  6,  7,  7,
    7,  7,  7,  8,  8,  8,  8,  9,  9,  9,  9, 10, 10, 10, 10, 11,
   11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 15, 15, 16, 16,
   16, 17, 17, 17, 18, 18, 18, 19, 19, 20, 20, 21, 21, 21, 22, 22,
   23, 23, 24, 24, 24, 25, 25, 26, 26, 27, 27, 28, 28, 29, 29, 30,
   30, 31, 32, 32, 33, 33, 34, 34, 35, 35, 36, 37, 37, 38, 38, 39,
   40, 40, 41, 41, 42, 43, 43, 44, 45, 45, 46, 47, 47, 48, 49, 50,
   50, 51, 52, 52, 53, 54, 55, 55, 56, 57, 58, 58, 59, 60, 61, 62,
   62, 63, 64, 65, 66, 67, 67, 68, 69, 70, 71, 72, 73, 74, 74, 75,
   76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91,
   92, 93, 94, 95, 96, 97, 98, 99,100,101,102,104,105,106,107,108,
  109,110,111,113,114,115,116,117,118,120,121,122,123,125,126,127
};

void initLEDStrip( uint16_t inNumLEDs ) {
  numLEDs = inNumLEDs;
  
  // Free the old buffer, if it exists
  if( buffer ) free( buffer );
  
  // Check out this tricky stuff: Pad the ending with 3+ empty (zero) bytes,
  // which tells the LPD8806 to latch after we push all the color data out.
  uint8_t latchBytes = (numLEDs + 31) / 32;

  bufferLen = (numLEDs+latchBytes) * 3;
  buffer = (uint8_t*)calloc( bufferLen, sizeof(uint8_t) );
  
  // Send 3+ zero-bytes to latch and begin receiving color data
  uint8_t i;
  for( i=0; i<latchBytes*3; i++ ) {
    //digitalWrite( clockPin, HIGH );
    //digitalWrite( clockPin, LOW );
    // Faster way to run the clock:
    ROM_GPIOPinWrite( GPIO_PORTB_BASE, CLOCK_PIN, CLOCK_PIN );  // clock high
    ROM_GPIOPinWrite( GPIO_PORTB_BASE, CLOCK_PIN, 0 );  // clock low
  }
}

void showStrip(void) {
  // Step through the buffer, push the whole thing to the strip
  uint16_t i;
  for( i=0; i<bufferLen; i++ ) {
    uint8_t aByte = buffer[i];
    
    // MSB first
    uint8_t b;
    for( b=7; ; b-- ) {
      //digitalWrite( dataPin, (brg>>b) & 0x1 );
      // Optimal, bitmask "& 0x1" not needed, because the bitmask 0x1 is the same as DATA_PIN (GPIO_PIN_0)
      ROM_GPIOPinWrite( GPIO_PORTB_BASE, DATA_PIN, (aByte>>b) /*& 0x1*/ );    // faster

      ROM_GPIOPinWrite( GPIO_PORTB_BASE, CLOCK_PIN, CLOCK_PIN );  // clock high
      ROM_GPIOPinWrite( GPIO_PORTB_BASE, CLOCK_PIN, 0 );  // clock low
      
      if( b==0 ) break;
    }
  }
  
}

// Valid color ranges: 0x0...0x7f
void setRGB( uint16_t idx, uint8_t r, uint8_t g, uint8_t b ) {
  uint16_t offset = idx*3;
  buffer[offset] = b | 0x80;
  buffer[offset+1] = r | 0x80;
  buffer[offset+2] = g | 0x80;
}

// Hue: 0.0..1.0 is a full cycle from red (0.0), through green (0.333), blue, (0.667), back to red (1.0).
// Sat: 0.0..1.0
// Val: 0.0..1.0
// WARNING: Out-of-bounds values of Sat and Val may produce unexpected behavior.
void setHSVfloats( uint16_t idx, float h, float s, float v, boolean useGammaCorrection ) {
  // Negative hues: Make sure these work as expected (make them positive before fmodf)
  if( h < 0.0 ) h -= floorf(h);
  
  // Convert h,s,v to RGB
  // Expand h by 6 times for the 6 color regions. Wrap the remainder to 0..1
  float hSix = fmodf( h*6.0, 6.0 );    // 0.0...6.0
  float hDec = hSix - floorf( hSix );  // decimal remainder
  
  // float v = v;                        // top (max) component
  float p = v * (1.0 - s);               // bottom (min) component
  float q = v * (1.0-(hDec*s));          // falling component (yellow->green: the red channel falls)
  float t = v * (1.0 - ((1.0-hDec)*s));  // rising component (red->yellow: the green channel rises)
  
  // prepare for RGB output
  float r, g, b;
  
  if( hSix < 1.0 )      { r=v; g=t; b=p; }  // red -> yellow
  else if( hSix < 2.0 ) { r=q; g=v; b=p; }  // yellow -> green
  else if( hSix < 3.0 ) { r=p; g=v; b=t; }  // green -> cyan
  else if( hSix < 4.0 ) { r=p; g=q; b=v; }  // cyan -> blue
  else if( hSix < 5.0 ) { r=t; g=p; b=v; }  // blue -> magenta
  else {                  r=v; g=p; b=q; }  // magenta -> red
  
  if( useGammaCorrection ) {
    setRGB( idx, gammaTable[(uint8_t)(r*0xff)], gammaTable[(uint8_t)(g*0xff)], gammaTable[(uint8_t)(b*0xff)] );
  } else {
    setRGB( idx, (uint8_t)(r*0x7f), (uint8_t)(g*0x7f), (uint8_t)(b*0x7f) );
  }
}

